package org.jocl.samples;

import static org.jocl.CL.*;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.nio.charset.Charset;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import org.jocl.*;

public class JOCLMatMul {
	
	String programSource;

	public JOCLMatMul() {
		//flops = mp(2n-1) ~ 2mnp ~ 2n^3
		
        // Create input- and output data 
        int N = 1000;
        float srcArrayA[] = new float[N*N];
        float srcArrayB[] = new float[N*N];
        float dstArray[] = new float[N*N];
        for (int i=0; i<N; i++) {
        	for (int j=0; j<N; j++) {
                srcArrayA[j*N + i] = (float) Math.random() * 1000.0f;
                srcArrayB[j*N + i] = (float) Math.random() * 1000.0f;        		
        	}
        }        
        Pointer pN = Pointer.to(new int[]{N});
        Pointer srcA = Pointer.to(srcArrayA);
        Pointer srcB = Pointer.to(srcArrayB);
        Pointer dst = Pointer.to(dstArray);

        // The platform, device type and device number
        // that will be used
        final int platformIndex = 0;
        final long deviceType = CL_DEVICE_TYPE_ALL;
        final int deviceIndex = 0;

        // Enable exceptions and subsequently omit error checks in this sample
        CL.setExceptionsEnabled(true);

        // Obtain the number of platforms
        int numPlatformsArray[] = new int[1];
        clGetPlatformIDs(0, null, numPlatformsArray);
        int numPlatforms = numPlatformsArray[0];

        // Obtain a platform ID
        cl_platform_id platforms[] = new cl_platform_id[numPlatforms];
        clGetPlatformIDs(platforms.length, platforms, null);
        cl_platform_id platform = platforms[platformIndex];

        // Initialize the context properties
        cl_context_properties contextProperties = new cl_context_properties();
        contextProperties.addProperty(CL_CONTEXT_PLATFORM, platform);
        
        // Obtain the number of devices for the platform
        int numDevicesArray[] = new int[1];
        clGetDeviceIDs(platform, deviceType, 0, null, numDevicesArray);
        int numDevices = numDevicesArray[0];
        
        // Obtain a device ID 
        cl_device_id devices[] = new cl_device_id[numDevices];
        clGetDeviceIDs(platform, deviceType, numDevices, devices, null);
        cl_device_id device = devices[deviceIndex];

        // Create a context for the selected device
        cl_context context = clCreateContext(
            contextProperties, 1, new cl_device_id[]{device}, 
            null, null, null);
        
        // Create a command-queue for the selected device
        cl_command_queue commandQueue = 
            clCreateCommandQueue(context, device, 0, null);

        // Allocate the memory objects for the input- and output data
        cl_mem memObjects[] = new cl_mem[3];
        memObjects[0] = clCreateBuffer(context, 
            CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR,
            Sizeof.cl_float * N * N, srcA, null);
        memObjects[1] = clCreateBuffer(context, 
            CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR,
            Sizeof.cl_float * N * N, srcB, null);
        memObjects[2] = clCreateBuffer(context, 
            CL_MEM_READ_WRITE, 
            Sizeof.cl_float * N * N, null, null);

        programSource = readfile("matmul.cl");
        // Create the program from the source code
        cl_program program = clCreateProgramWithSource(context,
            1, new String[]{ programSource }, null, null);
        
        // Build the program
        clBuildProgram(program, 0, null, null, null, null);
        
        // Create the kernel
        cl_kernel kernel = clCreateKernel(program, "mat_mul", null);
        
        // Set the arguments for the kernel
        clSetKernelArg(kernel, 0, Sizeof.cl_int, pN);
        clSetKernelArg(kernel, 1, 
            Sizeof.cl_mem, Pointer.to(memObjects[0]));
        clSetKernelArg(kernel, 2, 
            Sizeof.cl_mem, Pointer.to(memObjects[1]));
        clSetKernelArg(kernel, 3, 
            Sizeof.cl_mem, Pointer.to(memObjects[2]));
        
        // Set the work-item dimensions
        long global_work_size[] = new long[]{N,N};
        long local_work_size[] = new long[]{2,2};
                
        long start = System.nanoTime();
        // Execute the kernel
        //clEnqueueNDRangeKernel(commandQueue, kernel, 2, null,
        //    global_work_size, local_work_size, 0, null, null);
        clEnqueueNDRangeKernel(commandQueue, kernel, 2, null,
            global_work_size, null, 0, null, null);
        
        // Read the output data
        clEnqueueReadBuffer(commandQueue, memObjects[2], CL_TRUE, 0,
            N * N * Sizeof.cl_float, dst, 0, null, null);
        long dur = System.nanoTime() - start;
        long m = (long) N;
        long flops = m * m * (2*m - 1);
        double gflops = (double) flops/ (double) dur;
        
        NumberFormat format = NumberFormat.getInstance();
        format.setGroupingUsed(true);
        System.out.println("Matrix size: " + N);
        System.out.println("OpenCL:");
        System.out.println("Flops:" + format.format(flops));
        System.out.println("Dur:" + format.format(dur) + " nsecs");
        System.out.println("GFlops :" + format.format(gflops));
        
        // Release kernel, program, and memory objects
        clReleaseMemObject(memObjects[0]);
        clReleaseMemObject(memObjects[1]);
        clReleaseMemObject(memObjects[2]);
        clReleaseKernel(kernel);
        clReleaseProgram(program);
        clReleaseCommandQueue(commandQueue);
        clReleaseContext(context);        
        
        // Verify the result
//        boolean passed = true;
//        final float epsilon = 1e-7f;
//        for (int i=0; i<N; i++)
//        {
//            float x = dstArray[i];
//            float y = srcArrayA[i] * srcArrayB[i];
//            boolean epsilonEqual = Math.abs(x - y) <= epsilon * Math.abs(x);
//            if (!epsilonEqual)
//            {
//                passed = false;
//                break;
//            }
//        }
//        System.out.println("Test "+(passed?"PASSED":"FAILED"));
        if (N <= 10)
        {
        	System.out.println("Result: "+java.util.Arrays.toString(srcArrayA));
        	System.out.println("Result: "+java.util.Arrays.toString(srcArrayB));
            System.out.println("Result: "+java.util.Arrays.toString(dstArray));
        }		
        
        start = System.nanoTime();
        matmul(N, srcArrayA, srcArrayB, dstArray);
        dur = System.nanoTime() - start;
        m = (long) N;
        flops = m * m * (2*m - 1);        
        double gflops1 = (double) flops/ (double) dur;

        System.out.println("Java:");
        System.out.println("Flops:" + format.format(flops));
        System.out.println("Dur:" + format.format(dur) + " nsecs");
        System.out.println("GFlops :" + format.format(gflops1));
        System.out.println("Speedup : " + format.format(gflops / gflops1) + " times");
        
        if (N <= 10)
        {
        	System.out.println("Result: "+java.util.Arrays.toString(srcArrayA));
        	System.out.println("Result: "+java.util.Arrays.toString(srcArrayB));
            System.out.println("Result: "+java.util.Arrays.toString(dstArray));
        }		

	}
	
	String readfile(String filename) {
		StringBuffer sb = new StringBuffer(4096);
		byte buf[] = new byte[4096];
		try {			
			BufferedInputStream in = new BufferedInputStream(this.getClass().getResourceAsStream(filename),4096);
			while(true) {
				int l = in.read(buf);
				if (l == -1) break;
				sb.append(new String(buf, Charset.defaultCharset()),0,l);
			}
			in.close();
			return sb.toString();
		} catch (IOException e) {
			return "";
		}
	}
	
	void matmul(int N, float[] arrA, float[] ArrB, float[] ArrC) {
		ExecutorService service = Executors.newFixedThreadPool(8);
		
		int wg = 8;
		if(N < 8) 
			wg = N;
		else {
			while(N % wg != 0) {
				wg /= 2;
				if(wg == 1) break;
			}
		}
		
		ArrayList<Callable<Integer>> tasks = new ArrayList<Callable<Integer>>(8);
		for(int j=0; j<N; j++) {
			for(int i=0; i<N; i+=wg) {
				tasks.clear();
				tasks = new ArrayList<Callable<Integer>>(8);
				for(int ii=i; ii<i+wg; ii++) {
					MatMul task = new MatMul(N, arrA, ArrB, ArrC, ii, j);
					tasks.add(task);					
				}
				try {
					service.invokeAll(tasks);
				} catch (InterruptedException e) {
					//e.printStackTrace();
				}				
			}
		}
		service.shutdown();
	}

	
	public static void main(String[] args) {
		new JOCLMatMul();
	}
	
}
